{----------------------------------------------------------------------}
{ Scribe Source Code - version 1.6.0                                   }
{ written in QuickBasic 7.1                                            }
{ Translated to Dev-Pascal 1.9                                         }
{ Translated to Delphi 6                                               }
{----------------------------------------------------------------------}
Program Scribe;

{------------------- Program Notes ------------------------------------}
{ Scribe's coding is innefficient, very innefficient. Too much work is }
{ done with strings when enumerations could have been used. Some of the}
{ conditions are ridiculously long when arrays could have been used.   }
{ And objects (and methods) *should* be used instead of the multitude  }
{ of arrays and functions for them.                                    }
{                                                                      }
{ But, this is only an adapted version of the Scribe that was written  }
{ in QuickBASIC. That language didn't have any of the features         }
{ mentioned above. And instead of updating Scribe again with these     }
{ changes, I'm going to let it RIP. Later, Scribe will be completely   }
{ rewritten in Delphi or C++ for some sort of Daedalean IDE. But don't }
{ expect it anytime soon...                                            }
{----------------------------------------------------------------------}
{ All variables that can be passed by reference are passed by reference}
{    to save memory. This may help because Scribe is very recursive.   }
{ The keyWord array begins at 0 instead of 1 like the other arrays.    }
{    This is done so that keyWord[] can be passed to AnsiIndexText by  }
{    reference. Interestingly enough, when an array is passed this way,}
{    its starting index number becomes zero.                           }
{ The two global dummies, gd1 and gd2, are used so that I can have some}
{     procedures' params passed by reference. When the parameters are  }
{     not used in a call, a global dummy is used because a reference is}
{     demanded. Just a quick hack...                                   }
{------------------- Declare Shared Variables -------------------------}

uses SysUtils, Forms, Dialogs, StrUtils;

TYPE
	SymbolRec = Record
		stype, name, initVal : string;
		exten : array[1..10] of string;
		extCount : integer;
		extVal : array[1..5] of string;
	End;

VAR
	keyWord : array[0..500] of string;
	symbol : array[1..200] of SymbolRec;
	comment : array[1..100] of string;
	codeLine, codeComLine, symComLine,
	outputFile : string;
	file1, file2 : text;
	gd1, gd2 : integer; {global dummies}
        msgBox : TForm;

	varCount, numComs, longName, numKeys,
	blockLvl, makeBackup, lineNum, breakBrack,
	brackBreak, breakEndBrack, multiStat,
	lineLen, maxLineLen, endBrackbreak,
	indentBlocks, indentMsgs, msgBreak,
	indentReturn, breakReturn, addSymComLines,
	addCodeComLines, breakSymbols, maxExpLen : integer;

{--------------- Main Procedures --------------------------------------}
Procedure OpenCogs; Forward;
Procedure CloseCogs; Forward;
Procedure OutputSymbols; Forward;
Procedure ParseSymbols; Forward;
Procedure ParseCode; Forward;
Procedure ParseDesc; Forward;
Procedure ReadFileData; Forward;
Procedure ParseFlags; Forward;

{--------------- Parsing Subs -----------------------------------------}
Procedure GetToNextChar (Var codeLine:string; Var x:integer); Forward;
Function NextCharInLine (Var parseString:string; x:integer) : string; Forward;
Function CutAtNextDelimit (Var codeLine:string; Var x:integer; del:string) : string; Forward;
Procedure FindNextChar (Var codeLine:string; Var x:integer); Forward;
Procedure FindNextCharOrCom (Var codeLine:string; Var x:integer; eatCom:integer); Forward;
Procedure InputLn(Var x:integer); Forward;
Procedure Output(codeWord:string); Forward;
Procedure LnOutput(codeLine:string); Forward;
Procedure OutputLn(codeLine:string); Forward;

{--------------- String Functions -------------------------------------}
Function CharNum (Var codeLine:string; char:string) : integer; Forward;
Function IsCommentLine (Var codeLine:string) : integer; Forward;
Function CheckForBlankLine (Var codeLine:string; x:integer) : integer; Forward;

{--------------- Validation Functions ---------------------------------}
Function IsValidNum (num:string) : integer; Forward;
Function IsValidName (Var name:string) : integer; Forward;
Function IsOperator (Var char:string) : integer; Forward;
Function IsUnaryOp (Var char:string) : integer; Forward;

{--------------- Code Subs --------------------------------------------}
Procedure DoComment (Var codeLine:string; Var x:integer); Forward;
Procedure DoFor (Var codeLine:string; Var x:integer); Forward;
Procedure DoForStatement (Var codeLine:string; Var x:integer); Forward;
Procedure DoDo (Var codeLine:string; Var x:integer); Forward;
Procedure DoWhile (Var codeLine:string; Var x:integer); Forward;
Procedure DoStatement (Var codeLine:string; Var x, blockEnd, MsgReturn:integer; logical:integer); Forward;
Procedure DoIf (Var codeLine:string; Var x:integer); Forward;
Procedure DoElse (Var codeLine:string; Var x:integer); Forward;
Procedure DoBlock (Var codeLine:string; Var x:integer); Forward;
Procedure DoCall (Var codeLine:string; Var x:integer); Forward;
Procedure DoAssignment (Var codeLine, varName:string; Var x:integer; del:string); Forward;
Procedure DoMessage (Var codeLine:string; Var x:integer); Forward;
Procedure DoNextCode(Var codeLine:string; Var x:integer); Forward;
Procedure DoFunction (Var codeLine, funcName:string; Var x:integer); Forward;
Procedure DoString (Var codeLine:string; Var x:integer); Forward;
Procedure DoSubExp (Var codeLine:string; Var x:integer; del:string); Forward;
Procedure DoValue (Var codeLine:string; Var x:integer); Forward;
Procedure DoVector (Var codeLine:string; Var x:integer); Forward;
Procedure DoSemi (Var codeLine:string; Var x:integer); Forward;
Procedure DoReturn (Var codeWord, codeLine:string; Var x:integer; logical:integer); Forward;

{--------------- Misc Subs --------------------------------------------}
Procedure StoreComment (codeLine:string); Forward;
Function LookForComments (Var codeLine:string; Var x:integer) : integer; Forward;
Procedure DumpComments (desc:string); Forward;

{--------------- Prcedure Definition ----------------------------------}

Function CharNum (Var codeLine:string; char:string) : integer;
Var
	x, count : integer;
Begin
	count := 0;
	FOR x := 1 TO Length(codeLine) DO IF (Pos(Copy(codeLine, x, 1), char) > 0) THEN count := count + 1;
	CharNum := count;
END;

Function CheckForBlankLine (Var codeLine:string; x:integer) : integer;
	{-------------------------------------------------------------------}
	{ This function starts at the given position in a string and        }
	{ looks for non-space characters. It returns 1 for a blank line and }
	{ 0 if it finds a non-space character.                              }
	{-------------------------------------------------------------------}
Var
	char : string;
	blank : integer;
Begin
	blank := 1;
	FOR x := x TO Length(codeLine) DO
	Begin
		char := Copy(codeLine, x, 1);
		IF (char <> ' ') AND (char <> CHR(9)) THEN Begin
			blank := 0;
			break;
		END;
	End;
	CheckForBlankLine := blank;
END;

Procedure CloseCogs;
Var
	temp : text;
	fileName : string;
Begin
	CloseFile(file1);
	CloseFile(file2);

	If(makeBackup = 0) Then Begin
		AssignFile(temp, outputFile);
		Erase(temp);
	End
	Else Begin
		AssignFile(temp, outputFile);
		fileName := ExtractFilePath(outputFile) + ExtractFileName(outputFile) + '.bak';
		if(FileExists(fileName)) Then DeleteFile(fileName);
		Rename(temp, fileName);
	End;
	AssignFile(temp, ExtractFilePath(ParamStr(1)) + ExtractFileName(ParamStr(1)) + '.tmp');
	Rename(temp, outputFile);
End;

Function CutAtNextDelimit (Var codeLine:string; Var x:integer; del:string) : string;
Var
	lastBreak, x2 : integer;
	char : string;
Begin
	GetToNextChar(codeLine, x);
	lastBreak := x;

	For x2 := x To Length(codeLine) + 1 Do Begin
		char := Copy(codeLine, x2, 1);
		If (Pos(char, del) > 0) Then break;
		If ((Pos(' ', del) > 0) And (char = CHR(9))) Then break;
	End;

	x := x2;
	CutAtNextDelimit := Trim(Copy(codeLine, lastBreak, x - lastBreak));
End;

Procedure DoAssignment (Var codeLine, varName:string; Var x:integer; del:string);
Var
	varNum : integer;
Begin
	For varNum := 1 To varCount + 1 Do If (LowerCase(symbol[varNum].name) = LowerCase(varName)) Then break;
	If (varNum <= varCount) Then Output(symbol[varNum].name)
	Else Output(LowerCase(varName));

	FindNextChar(codeLine, x);
	If (Copy(codeLine, x, 1) = '[') Then Begin
		Output('[');
		x := x + 1;
		DoSubExp(codeLine, x, ']');     {process the expression.}
		Output(']');
		FindNextChar(codeLine, x);
	End;

	x := x + 1;
	if((lineLen + 3 < maxExpLen)) Then Output(' = ')
	else Begin Output(' ='); lineLen := maxLineLen; End;
	DoSubExp(codeLine, x, del);      {note that the calling sub chooses the delimit that the assignment should end with.}
End;

Procedure DoBlock (Var codeLine:string; Var x:integer);
Var
	blockEnd : integer;
Begin
	blockEnd := 0;
	x := x + 1;     {get past the '{'}

	// If the line has not been broken and it should be, break it.
	If(((lineLen > 0) and (breakBrack = 1)) Or (lineLen >= maxLineLen)) Then OutputLn('');

	// If the line should be broken, output an indent; else, output a space.
	if((breakBrack = 1) or (lineLen = 0)) Then Output(StringOfChar(Chr(9), blockLvl*indentBlocks + indentMsgs) + '{')
	else if(lineLen > 0) Then Output(' {')
	else Output('{');

	if(brackBreak = 1) Then OutputLn('');
	blockLvl := blockLvl + 1;

	While Not EOF(file1) Do Begin
		WHILE ((x <= Length(codeLine)) AND (CheckForBlankLine(codeLine, x) = 0) AND (blockEnd = 0)) Do
			DoStatement(codeLine, x, blockEnd, gd1, 0);
   		if(blockEnd = 1) Then break;
		InputLn(x);
	End;

	blockLvl := blockLvl - 1;

	// If the line has not been broken and it should be, break it.
	If((lineLen > 0) and (breakEndBrack = 1)) Then OutputLn('');

	// If the line should be broken (or has been via comment), output an indent; else, output a space.
	if((breakEndBrack = 1) OR (lineLen = 0)) Then Output(StringOfChar(Chr(9), blockLvl*indentBlocks + indentMsgs))
	else if(lineLen > 0) Then Output(' ');

	If(endBrackBreak = 1) Then OutputLn('}')
	else Output('}');
END;

Procedure DoCall (Var codeLine:string; Var x:integer);
Var
	msg : string;
Begin
	msg := LowerCase(CutAtNextDelimit(codeLine, x, ' ;#/'));
	Output('call ' + msg);
	DoSemi(codeLine, x);
END;

Procedure DoComment (Var codeLine:string; Var x:integer);
Begin
	Output(Trim(Copy(codeLine, x, Length(codeLine) - x + 1)));
	lineLen := maxLineLen;
	InputLn(x);
END;

Procedure DoDo (Var codeLine:string; Var x:integer);
Begin
	Output('do');
	DoNextCode(codeLine, x);  { notice that we don't need to check for the while statement. }
END;                              { We'll let DoWhile take care of it. DoNextCode() will allow  }
                                  { the while(); to end with a semicolon.                       }
Procedure DoFor(Var codeLine:string; Var x:integer);
Begin
	{------------ Check the variable init -------------------------}

	Output('for(');
	FindNextChar(codeLine, x); {get to the paren}
	x := x + 1;                {get past open paren}
        FindNextChar(codeLine, x); {paren may be on the next line.}
        If(NextCharInLine(codeLine, x) <> ';') Then Begin
                Repeat Begin
                        FindNextChar(codeLine, x);
                        DoForStatement(codeLine, x);
                        FindNextChar(codeLine, x);
                        If(NextCharInLine(codeLine, x) <> ';') Then Output(', ');
                End; Until((NextCharInLine(codeLine, x) = ';') OR (EOF(file1)));
        End;
	x := x + 1;     {get past the ';' delimiter}
	Output('; ');

	{------------ Check the condition -----------------------------}

        FindNextChar(codeLine, x);
        If(NextCharInLine(codeLine, x) <> ';') Then Begin
        	DoSubExp(codeLine, x, ';');  {DoSubExp uses FindNextChar; also leaves x at the semicolon}
                x := x + 1;
        End
        Else x := x + 1;
	Output('; ');

	{------------ Check the incrementation ------------------------}

        Repeat Begin
        	FindNextChar(codeLine, x);
                DoForStatement(codeLine, x);
                FindNextChar(codeLine, x);
                If(NextCharInLine(codeLine, x) <> ')') Then Output(', ');
        End; Until((NextCharInLine(codeLine, x) = ')') OR (EOF(file1)));

        x := x + 1;
	Output(')');

        {------------ Check the following code ------------------------}

	DoNextCode(codeLine, x);
END;

Procedure DoForStatement(Var codeLine:string; Var x:integer);
Var
	nextChar, codeWord : string;
Begin
	codeWord := LowerCase(CutAtNextDelimit(codeLine, x, ' =,();{}[:/#'));
        FindNextCharOrCom(codeLine, x, 1);
	nextChar := NextCharInLine(codeLine, x);

	If((nextChar = '(') AND (Length(codeWord) > 0) {AND (AnsiIndexText(codeWord, keyword) > -1)} AND NOT (IsValidNum(codeWord) = 1)) Then
		DoFunction(codeLine, codeWord, x)

	ELSE If(IsValidName(codeWord) = 1) Then Begin
		DoAssignment(codeLine, codeWord, x, ',;)');
                // DoSubExp (called by DoAssignment) will skip a closing ')'.
                // However, DoFor (the caller) will need to verify that delimiter,
                // so we'll go back one here. NEVER have a low-level function treat
                // delimiters differently!
                If(NextCharInLine(codeLine, x - 1) = ')') Then x := x - 1;
	End
	ELSE Begin
                msgBox := CreateMessageDialog('Error in for sub-statement syntax on line ' + IntToStr(lineNum) + '. Aborting.', mtError, [mbOK]);
                msgBox.Caption := 'Scribe 1.6.0';
                msgBox.ShowModal;
		CloseFile(file1);
		CloseFile(file2);
		Erase(file2);
		Halt;
	END;
END;

Procedure DoFunction(Var codeLine, funcName:string; Var x:integer);
Var
        nextChar : string;
	verbNum : integer;
Begin
	{------------ Find the function's name ------------------------}

	verbNum := AnsiIndexText(funcName, keyword);
	If(verbNum > -1) Then Output(keyWord[verbNum] + '(')
	ELSE Output(UpperCase(funcName) + '(');

	{------------ Parse function's arguments ----------------------}

        FindNextChar(codeLine, x);
	x := x + 1;        {get past begin paren.}
	If(NextCharInLine(codeLine, x) = ')') Then Begin
		x := x + 1;
		Output(')');
	End
	ELSE Begin
		Repeat
			DoSubExp(codeLine, x, ',)');
			nextChar := Copy(codeLine, x - 1, 1);
			If(nextChar = ',') Then Begin {nextChar may never be ')', but if it isn't, DoSubExp() will catch the error.}
				if(lineLen + 2 < maxExpLen) Then Output(', ')
				else Begin Output(','); lineLen := maxExpLen; End;
			End;
		Until(nextChar = ')');
		Output(')');
	END;
END;

Procedure DoIf(Var codeLine:string; Var x:integer);
Begin
	Output('if(');
	FindNextChar(codeLine, x);
	x := x + 1;     {get past the '('}
	DoSubExp(codeLine, x, ')');
	Output(')');
	DoNextCode(codeLine, x);
END;

Procedure DoElse(Var codeLine:string; Var x:integer);
Begin
	Output('else');
	DoNextCode(codeLine, x);
END;

Procedure DoMessage(Var codeLine:string; Var x:integer);
Var
	MsgReturn : integer;
Begin
	MsgReturn := 0;
	If(msgBreak = 1) Then OutputLn('');

	WHILE Not EOF(file1) Do Begin
		WHILE ((x <= Length(codeLine)) AND (MsgReturn = 0) AND (CheckForBlankLine(codeLine, x) = 0)) Do
			DoStatement(codeLine, x, gd1, MsgReturn, 0);
  		If(MsgReturn = 1) Then break;
		InputLn(x);
        End;
END;

Procedure DoNextCode(Var codeLine:string; Var x:integer);
Var
	char : string;
Begin

	FindNextChar(codeLine, x);
	char := Copy(codeLine, x, 1);

	If(char = ';') Then Begin	{ Note here that all conditionals are allowed to end in semicolons. }
		x := x + 1;      {get past the semicolon}
		If(LookForComments(codeLine, x) = 0) Then
			Output(';')
		ELSE Begin
			Output(';');
			DumpComments('firstend');
                        OutputLn('');
             		InputLn(x);
		End;
	END
	ELSE If(char = '{') Then Begin
		If(numComs > 0) Then Begin
                        DumpComments('firstend');
                        OutputLn('');
                End;
		DoBlock(codeLine, x);
	End
	ELSE Begin
		If(numComs > 0) Then Begin
			DumpComments('firstend');
                        OutputLn('');
			DoStatement(codeLine, x, gd1, gd2, 1);
		End
		ELSE Begin
			DoStatement(codeLine, x, gd1, gd2, 1);
		END;
	END;
End;

Procedure DoReturn(Var codeWord, codeLine:string; Var x:integer; logical:integer);
Begin
	if((blockLvl = 0) and (logical = 0)) Then Begin
		if(breakReturn = 0) Then Begin
			if((multiStat = 0) or (lineLen >= maxLineLen)) Then OutputLn('')
			else if(lineLen > 0) Then Output(' ');
		End
		else Begin
			// Remember here that the line may already be broken... or it may not be. :(
			if((breakReturn >= 1) and (lineLen > 0)) Then OutputLn('');
			if(breakReturn = 2) Then OutputLn('');
			if((indentReturn = 1) and ((multiStat = 0) OR (breakReturn > 0))) Then Output(StringOfChar(Chr(9), 1));
		End;
	End;

	If(codeWord = 'return') Then Output('Return')
	ELSE If(codeWord = 'stop') Then Output('Stop');

	DoSemi(codeLine, x);
	if((blockLvl = 0) and (logical = 0)) Then OutputLn('');
END;

Procedure DoSemi(Var codeLine:string; Var x:integer);
Begin
	FindNextChar(codeLine, x);       {get to the semicolon}
	x := x + 1;                      {get past the semicolon}
	If(LookForComments(codeLine, x) = 1) Then Begin
		Output(';');
		DumpComments('firstend');
		InputLn(x);
	End
	ELSE Output(';');
END;

Procedure DoStatement(Var codeLine:string; Var x, blockEnd, MsgReturn:integer; logical:integer);
Var
	nextChar, codeWord : string;
Begin
	codeWord := LowerCase(CutAtNextDelimit(codeLine, x, ' =();{}[:/#'));
        If(Length(codeWord) > 0) Then FindNextCharOrCom(codeLine, x, 1) //eat comments! :(]
        Else FindNextCharOrCom(codeLine, x, 0); // else if we think we found a comment, don't eat it
	nextChar := NextCharInLine(codeLine, x);

	{------------------------ Handle Indentation ----------------------------------------}

	if(((codeWord <> 'return') And (codeWord <> 'stop')) OR (blockLvl > 0) OR (logical = 1)) Then Begin
		if((lineLen >= maxLineLen) And (multiStat = 1)) Then OutputLn('')
		else if((multiStat = 0) and (lineLen > 0) and (nextChar <> '}') and (logical = 0)) Then Begin
			// In the case of 'else{' nextChar will be a '{'. So we have to add this condition:
			if(((nextChar = '{') and (Length(codeWord) > 0)) OR (nextChar <> '{')) Then OutputLn('');
		End;

		if(((length(codeWord) > 0) OR (nextChar[1] in ['/','#']))) Then Begin
			If(lineLen = 0) Then Output(StringOfChar(Chr(9), blockLvl*indentBlocks + indentMsgs))
			else Output(' ');
		End;
	End;

	If((codeWord = 'if') AND (nextChar = '(')) Then
		DoIf(codeLine, x)
	ELSE If(codeWord = 'else') Then
		DoElse(codeLine, x)
	ELSE If(codeWord = 'call') Then
		DoCall(codeLine, x)
	ELSE If(codeWord = 'do') Then
		DoDo(codeLine, x)
	ELSE If((codeWord = 'while') AND (nextChar = '(')) Then
		DoWhile(codeLine, x)
	ELSE If((codeWord = 'for') AND (nextChar = '(')) Then
		DoFor(codeLine, x)
	ELSE If((codeWord = 'return') OR (codeWord = 'stop')) Then Begin
		DoReturn(codeWord, codeLine, x, logical);
		MsgReturn := 1;
	End
	ELSE If(nextChar = '{') Then
		DoBlock(codeLine, x) {this code block is technically useless, though it is syntactically correct.}
                                     {In Parsec, we'd flak the person who wrote this. :p }
	ELSE If(nextChar = '}') Then Begin
		x := x + 1;     {get past the bracket}
		blockEnd := 1;
	End
	ELSE If((nextChar = '(') AND (Length(codeWord) > 0) {AND (AnsiIndexText(codeWord, keyword) > -1)} AND NOT (IsValidNum(codeWord) = 1)) Then Begin
		DoFunction(codeLine, codeWord, x);
		DoSemi(codeLine, x);
	End
	ELSE If(nextChar = ':') Then Begin
		Output(LowerCase(codeWord) + ':');
		x := x + 1;     {get past the ':'}
	End
	ELSE If((codeWord = '') AND ((nextChar = '#') OR (nextChar = '/'))) Then
		DoComment(codeLine, x)
	ELSE If((IsValidName(codeWord) = 1) AND ((nextChar = '=') OR (nextChar ='['))) Then Begin
		DoAssignment(codeLine, codeWord, x, ';');
		DoSemi(codeLine, x);
	End
	ELSE Begin
                msgBox := CreateMessageDialog('Error in statement syntax on line ' + IntToStr(lineNum) + '. Aborting.', mtError, [mbOK]);
                msgBox.Caption := 'Scribe 1.6.0';
                msgBox.ShowModal;
		CloseFile(file1);
		CloseFile(file2);
		Erase(file2);
		Halt;
	END;
	DumpComments('firstend');
END;

Procedure DoString(Var codeLine:string; Var x:integer);
Var
	char : string;
	x2 : integer;
Begin
	Output('"');

        // we use x2 because Delphi does not allow the use of a referenced variable
        // as a for loop counter.
	FOR x2 := x + 1 TO Length(codeLine) Do Begin	{Add 1 to x to get past opening quotes}
		char := Copy(codeLine, x2, 1);
		Output(char);
		If(char = '"') Then break;
	End;
	x := x2 + 1      {get past end quote}
END;

Procedure DoSubExp(Var codeLine:string; Var x:integer; del:string);
Var
	char, altChar, expect : string;
Begin
	expect := 'value';

	WHILE not EOF(file1) Do Begin
		FindNextChar(codeLine, x);
		char := Copy(codeLine, x, 1);
		altChar := Copy(codeLine, x, 2);

		If((lineLen >= maxExpLen) AND (Pos(char, del) = 0)) Then Begin
			if(numComs > 0) Then DumpComments('firstend')
			else OutputLn('');
			If(blockLvl > 0) Then Begin
				If(indentBlocks = 1) Then Output(StringOfChar(Chr(9), blockLvl + indentMsgs))
				Else Output(StringOfChar(Chr(9), indentMsgs));
			End
			else If(indentMsgs = 1) Then Output(StringOfChar(Chr(9), 1));
		End;

		If((IsUnaryOp(char) = 1) AND (expect = 'value')) Then Begin
			x := x + 1;
			Output(char);
		End
		ELSE If((expect = 'value') AND (char = '(')) Then Begin
			x := x + 1;
                        Output('(');
			DoSubExp(codeLine, x, ')');
                        Output(')');
			expect := 'op/end';
		End
		ELSE If(expect = 'value') Then Begin
			DoValue(codeLine, x);
			expect := 'op/end';
		End
		ELSE If((IsOperator(altChar) = 1) AND (expect = 'op/end')) Then Begin    {note that we check for two-char ops first.}
			expect := 'value';
			x := x + 2;     {get past second char of operator.}
			if((lineLen > 0) and (lineLen + 4 < maxExpLen)) Then Output(' ' + altChar + ' ')
			else if(lineLen + 4 >= maxExpLen) Then Begin Output(' ' + altChar); lineLen := maxLineLen; End
			else Output(altChar + ' ');
		End
		ELSE If((IsOperator(char) = 1) AND (expect = 'op/end')) Then Begin
			expect := 'value';
			x := x + 1;
			if((lineLen > 0) and (lineLen + 3 < maxExpLen)) Then Output(' ' + char + ' ')
			else if(lineLen + 3 >= maxExpLen) Then Begin Output(' ' + char); lineLen := maxLineLen; End
			else Output(char + ' ');
		End
		ELSE If(Pos(char, del) > 0) Then Begin
			{in most cases, the semicolon is checked by seperate subs, so we won't pass it here.}
			If(char <> ';') Then x := x + 1;     {get past the delimiter.}
			break;
		End
		ELSE Begin
                        msgBox := CreateMessageDialog('Error in expression on line ' + IntToStr(lineNum) + '. Aborting.', mtError, [mbOK]);
                        msgBox.Caption := 'Scribe 1.6.0';
                        msgBox.ShowModal;
			CloseFile(file1);
			CloseFile(file2);
			Erase(file2);
			Halt;
		END;
	End;
END;

Procedure DoValue(Var codeLine:string; Var x:integer);
Var
        varNum : integer;
	nextChar, codeWord : string;
Begin
	FindNextChar(codeLine, x);
	codeWord := CutAtNextDelimit(codeLine, x, ' ;()[],=!><&|%*/-+"#/' + Chr(39));
        FindNextCharOrCom(codeLine, x, 1);
	nextChar := NextCharInLine(codeLine, x);

	If(nextChar = '"') Then       {Strings can contain other delimiters so we can't use an IsValid function.}
		DoString(codeLine, x) {This is more efficient anyway...}

	ELSE If(nextChar = Chr(39)) Then    {Same as above.}
		DoVector(codeLine, x)

	ELSE If((nextChar = '(') AND (Length(codeWord) > 0) {AND (AnsiIndexText(codeWord, keyword) > -1)} AND NOT (IsValidNum(codeWord) = 1)) Then
                // we must be sure that we have a verb before calling this function.
                // if we did not check codeWord's length, then things like '1 (8)' would be treated as verbs!
		DoFunction(codeLine, codeWord, x)

	ELSE If(IsValidNum(codeWord) = 1) Then
		Output(codeWord)        // accepts decimal, flex, hex, and octal values.

	ELSE If(IsValidName(codeWord) = 1) Then Begin    {handles vars and message names.}
		FOR varNum := 1 TO varCount + 1 Do If(LowerCase(symbol[varNum].name) = LowerCase(codeWord)) Then break;
		If(varNum <= varCount) Then Output(symbol[varNum].name)
		ELSE Output(LowerCase(codeWord));

		If(nextChar = '[') Then Begin
			Output('[');
			x := x + 1;          {get past the '['}
			DoSubExp(codeLine, x, ']');  {evaluate the expression}
			Output(']');
		END;
	END;
END;

Procedure DoVector(Var codeLine:string; Var x:integer);
Var
	xvec, yvec, zvec : string;
Begin
	x := x + 1;     {get past opening quote}
	xvec := CutAtNextDelimit(codeLine, x, ' ');
	yvec := CutAtNextDelimit(codeLine, x, ' ');
	zvec := CutAtNextDelimit(codeLine, x, ' ' + Chr(39));
	x := x + 1;     {get past end quote}
	Output(Chr(39) + xvec + ' ' + yvec + ' ' + zvec + Chr(39));
END;

Procedure DoWhile(Var codeLine:string; Var x:integer);
Begin
	Output('while(');
	FindNextChar(codeLine, x);
	x := x + 1;     {get past the '('}
	DoSubExp(codeLine, x, ')');
	Output(')');
	DoNextCode(codeLine, x);
END;

Procedure DumpComments(desc:string);
Var
	a : integer;
Begin
	If(numComs > 0) Then Begin
		FOR a := 1 TO numComs Do Begin
			If(desc = 'stat') Then Output(StringOfChar(Chr(9), blockLvl*indentBlocks + indentMsgs) + comment[a])
			ELSE If(desc = 'firstend') Then Begin
				Output(StringOfChar(Chr(9), 1) + comment[a]);
				desc := 'stat';
			End
			Else If(desc = 'noindent') Then Output(comment[a])
			ELSE Output(StringOfChar(Chr(9), 1) + comment[a]);
			numComs := numComs - 1;
			if(numComs > 0) Then OutputLn('');
		End;
		// if multiStat is disabled, a new line will be created anyway.
		// but if enabled, DoStatement will create a new line because lineLen = maxLineLen.
		lineLen := maxLineLen;
	END;
END;

Procedure FindNextChar(Var codeLine:string; Var x:integer);
Begin
	WHILE Not EOF(file1) Do Begin
		If(CheckForBlankLine(codeLine, x) = 0) Then Begin
			GetToNextChar(codeLine, x);
			If(LookForComments(codeLine, x) = 0) Then break;
		END;
		InputLn(x);
	End;
END;

Procedure FindNextCharOrCom(Var codeLine:string; Var x:integer; eatCom:integer);
Begin
        // This is used by several functions: DoForStatement, DoStatement, and DoValue.
        // DoStatement is the only one to make use of the last param however.
        // The original FindNextChar() ate comments indiscriminantly. That means it would
        // eat comments at a statement level (and possibly overrun the comment array)
        // and DoComment() would be useless. So when DoStatement thinks it has found a
        // comment, it tells this function not to eat it.

	WHILE Not EOF(file1) Do Begin
		If(CheckForBlankLine(codeLine, x) = 0) Then Begin
			GetToNextChar(codeLine, x);
                        If(eatCom = 1) Then Begin
                                If(LookForComments(codeLine, x) = 0) Then break;
                        End
                        Else break;
		END;
		InputLn(x);
	End;
END;

Procedure GetToNextChar(Var codeLine:string; Var x:integer);
Var
	char : string;
	x2 : integer;
Begin
	FOR x2 := x TO Length(codeLine) Do Begin
		char := Copy(codeLine, x2, 1);
		If((char <> ' ') AND (char <> Chr(9))) Then break;
	End;
	x := x2;
END;

Procedure InputLn(Var x:integer);
Begin
	x := 1;
	ReadLn(file1, codeLine);
	lineNum := lineNum + 1;
End;

Function IsCommentLine(Var codeLine:string) : integer;
Var
	nextChar, next2Char : string;
Begin
	codeLine := Trim(codeLine);
	nextChar := Copy(codeLine, 1, 1);
	next2Char := Copy(codeLine, 1, 2);
	IsCommentLine := 0;

	If((nextChar = '#') OR (next2Char = '//')) Then
		If((CharNum(codeLine, ' #/.*_~=-') = Length(codeLine)) and (Length(codeLine) > 2)) Then IsCommentLine := 1;
END;

Function IsOperator(Var char:string) : integer;
Var valid : integer;
Begin
        // I know it's disgusting! I should have an array of operators, but I don't...
	if((char = '==') or (char = '>=') or (char = '<=') or (char = '&&') or (char = '||') or (char = '!=')
	or (char = '+') or (char = '-') or (char = '*') or (char = '/') or (char = '%') or (char = '>') or
	(char = '<') or (char = '&') or (char = '|')) THEN valid := 1
	else valid := 0;
	if(Length(char) < 1) Then valid := 0;
	IsOperator := valid;
END;

Function IsUnaryOp(Var char:string) : integer;
Begin
	If((char = '-') or (char = '!')) Then IsUnaryOp := 1
	ELSE IsUnaryOp := 0;
END;

Function IsValidNum(num:string) : integer;
Var
	isHex, isFlex, valid, x : integer;
	digit : char;
Begin
	valid := 1;
	isHex := 0;
	isFlex := 0;
	num := LowerCase(num);

	{ Find out if this is a flex or hex number. }
	If(LeftStr(num, 2) = '0x') Then Begin isHex := 1; num := RightStr(num, Length(num) - 2); END
	Else If(Pos('.', num) > 0) Then isFlex := 1;

	FOR x := 1 TO Length(num) Do Begin
		digit := Copy(num, x, 1)[1];
		If(isHex = 1) Then Begin If not(digit in ['0'..'9', 'a'..'f']) Then valid := 0; End
		Else If(isFlex = 1) Then Begin If not(digit in ['0'..'9', '.']) Then valid := 0; End
		ELSE If not(digit in ['0'..'9']) Then Begin valid := 0; End;
		if(valid = 0) Then break;
	End;

	If(isFlex = 1) Then If(CharNum(num, '.') > 1) Then valid := 0;
	If(CheckForBlankLine(num, 1) = 1) Then valid := 0;
	IsValidNum := valid;
END;

Function IsValidName(Var name:string) : integer;
Var
	x, valid, hasAlpha : integer;
	temp : string;
	chr : char;
Begin
	valid := 1;
	hasAlpha := 0;

	FOR x := 1 TO Length(name) Do Begin
		temp := LowerCase(Copy(name, x, 1));
		chr := temp[1];
		{if charint is not 0-9 or a lowercase letter, or an underscore, or a dash, then there's an error.}
		IF not(chr in ['0'..'9', 'a'..'z', '_', '-']) Then Begin
			valid := 0;
			break;
		END;
		IF(chr in ['a'..'z']) Then hasAlpha := 1;
	End;

	If(hasAlpha = 0) Then valid := 0
        // quick hack because Scribe does not recognize reserved words!
        else If((name = 'end') OR (name = 'code') OR (name = 'if') OR (name = 'do')
        OR (name = 'while') OR (name = 'for') OR (name = 'else') OR (name = 'call')
        OR (name = 'return') OR (name = 'stop')) Then valid := 0;
	IsValidName := valid;
END;

Function LookForComments(Var codeLine:string; Var x:integer) : integer;
Begin
	GetToNextChar(codeLine, x);
	If((Copy(codeLine, x, 1) = '#') OR (Copy(codeLine, x, 2) = '//')) Then Begin
		numComs := numComs + 1;
		comment[numComs] := Copy(codeLine, x, Length(codeLine) - x + 1);
		LookForComments := 1;
	End
	ELSE LookForComments := 0;
END;

Function NextCharInLine(Var parseString:string; x:integer) : string;
Var
	char, nextChar : string;
Begin
	FOR x := x TO Length(parseString) Do Begin
		char := Copy(parseString, x, 1);
		If((char <> ' ') AND (char <> Chr(9))) Then Begin
			nextChar := char;
			break;
		END;
	End;
	NextCharInLine := nextChar;
END;

Procedure OpenCogs;
Begin
        // Note here that I'm betting all of the file's path will be
        // in the first CDM argument.  So there must be a "%1" after
        // Scribe's path in the File Type Action settings.
        outputFile := ParamStr(1);
	if(Not FileExists(outputFile)) Then Begin
                msgBox := CreateMessageDialog('Input file not found. File path is invalid.', mtError, [mbOK]);
                msgBox.Caption := 'Scribe 1.6.0';
                msgBox.ShowModal;
		Halt;
	End;
	AssignFile(file1, outputFile);
	Reset(file1);
	AssignFile(file2, ExtractFilePath(ParamStr(1)) + ExtractFileName(ParamStr(1)) + '.tmp');
	Rewrite(file2);
END;

Procedure Output(codeWord:string);
Begin
	Write(file2, codeWord);
	lineLen := lineLen + Length(codeWord) - CharNum(codeWord, Chr(9));
End;

Procedure LnOutput(codeLine:string); { Create a new line and then output a string. }
Begin
	WriteLn(file2, '');
	lineLen := 0;
	Write(file2, codeLine);
	lineLen := lineLen + Length(codeLine) - CharNum(codeLine, Chr(9));
End;

Procedure OutputLn(codeLine:string); { Output a string and then create a new line. }
Begin
	WriteLn(file2, codeLine);
	lineLen := 0;
End;

Procedure OutputSymbols;
        { The beauty of this function is that it's type-dumb. It doesn't have to know }
        { what type of variable its outputting; it outputs all variables of one type, }
        { adds the type to its list of done types, and then goes to the next type.    }
        { Effectively, this allows you to use types that don't exist and outputs the  }
        { variables in order of their type.                                           }
Var
	outType, doneTypes : string;
	outVars, a, b, nameLen : integer;
Begin
	doneTypes := 'a';
	outType := 'a';
	outVars := 0;

	if(addSymComLines = 1) Then OutputLn(symComLine);
	OutputLn('symbols');
	if(breakSymbols = 1) Then OutputLn('');

	WHILE (outVars < varCount) Do Begin
		a := 1;
		WHILE((Pos(outType, doneTypes) > 0) AND (a <= varCount)) Do Begin
			outType := symbol[a].stype;
			a := a + 1;
		End;

		FOR a := 1 TO varCount Do Begin
			If(symbol[a].stype = outType) Then With symbol[a] Do Begin
				If(initVal = 'null') Then nameLen := Length(name)
				ELSE nameLen := Length(name) + Length(initVal) + 1;

				Output(LowerCase(stype));
				Output(StringOfChar(Chr(32), 10 - Length(stype)));
				Output(name);
				If(initVal <> 'null') Then Output('=' + initVal);
				If(extCount > 0) Then Output(StringOfChar(Chr(32), longName - nameLen + 2));
				FOR b := 1 TO extCount Do Begin
					Output(' ');
					Output(LowerCase(exten[b]));
					If(extVal[b] <> 'null') Then Output('=' + extVal[b]);
				End;
				OutputLn('');
				outVars := outVars + 1;
			END;
		End;
		doneTypes := doneTypes + outType;
		if(breakSymbols = 1) Then OutputLn('');
	End;

	If(numComs > 0) Then Begin
		DumpComments('noindent');
		OutputLn('');
		if(breakSymbols = 1) Then OutputLn('');
	End;

	OutputLn('end');
	If(addSymComLines = 1) Then OutputLn(symComLine);
END;

Procedure ParseCode;
Var
	x : integer;
	nextChar, numString, codeWord : string;
Begin
	OutputLn('code');
	if(addCodeComLines = 1) Then OutputLn(codeComLine);

	WHILE not(EOF(file1)) Do Begin
		InputLn(x);
		nextChar := NextCharInLine(codeLine, 1);
                codeWord := LowerCase(CutAtNextDelimit(codeLine, x, ':#/'));
		If((CheckForBlankLine(codeLine, 1) = 0) AND (nextChar <> '#') AND (nextChar <> '/') AND (codeWord <> 'code') AND (codeWord <> 'end')) Then Begin
			If(IsValidName(codeWord) = 0) Then Begin
				Str(lineNum, numString);
                                msgBox := CreateMessageDialog('Invalid message name on line ' + numString + '.', mtError, [mbOK]);
                                msgBox.Caption := 'Scribe 1.6.0';
                                msgBox.ShowModal;
				CloseFile(file1);
				CloseFile(file2);
				Erase(file2);
				Halt;
			END;
			x := x + 1;     {get past the ':'}
			Output(codeWord + ':');
			DumpComments('stat');
			DoMessage(codeLine, x);
			if(addCodeComLines = 1) Then OutputLn(codeComLine);
		End
		ELSE If((nextChar = '#') OR (nextChar = '/')) Then Begin
			If((IsCommentLine(codeLine) = 0) OR (addCodeComLines = 0)) Then OutputLn(codeLine);
		END;
	End;
	Output('end');
END;

Procedure ParseDesc;
Var
        codeWord : string;
Begin
	WHILE not EOF(file1) Do Begin
		InputLn(gd1);
		If(CheckForBlankLine(codeLine, 1) = 0) Then Begin
                        gd1 := 1;
                        codeWord := LowerCase(CutAtNextDelimit(codeLine, gd1, ' #/'));
			If((codeWord = 'symbols') OR (LeftStr(codeWord, 5) = 'flags') OR (codeWord = 'code')) Then break;
			If((IsCommentLine(codeLine) = 0) OR (addSymComLines = 0)) Then OutputLn(codeLine);
		END;
	End;
END;

Procedure ParseFlags;
Var
	x : integer;
	flagVal : string;
Begin
        x := 1;
	FindNextChar(codeLine, x);
        If(CutAtNextDelimit(codeLine, x, ' =') <> 'flags') Then Exit;
	FindNextChar(codeLine, x);      {get to the '='}
	x := x + 1;                     {get past the '='}
        FindNextChar(codeLine, x);      {get to the flag value}
	flagVal := Trim(Copy(codeLine, x, Length(codeLine) + 1 - x));
	if(addSymComLines = 1) Then OutputLn(symComLine);
	OutputLn('flags=' + flagVal);
	ParseDesc;
END;

Procedure ParseSymbols;
Var
	a, x, extCount : integer;
	stype, name, initVal, nextChar, codeWord : string;
	extVal, exten : array[1..5] of string;
Begin
	varCount := 0;

	WHILE not EOF(file1) Do Begin
		InputLn(x);
		nextChar := NextCharInLine(codeLine, 1);
                codeWord := LowerCase(CutAtNextDelimit(codeLine, x, ' #/'));
                x := 1;         {CAND() just changed this, so set it back.}
		If((CheckForBlankLine(codeLine, 1) = 0) AND (nextChar <> '#') AND (nextChar <> '/') AND (codeWord <> 'symbols')) Then Begin

			{---------- Check Line ------------------}

			If(codeWord = 'end') Then break;

			{---------- Reset Vars ------------------}

			stype := 'null';
			name := 'null';
			initVal := 'null';
			extCount := 0;
   			FOR a := 1 TO 5 Do Begin
				extVal[a] := 'null';
				exten[a] := 'null';
			End;

			{---------- Main Symbol Parser ----------}

			stype := CutAtNextDelimit(codeLine, x, ' ');
			name := CutAtNextDelimit(codeLine, x, ' =#/');

			GetToNextChar(codeLine, x);
			If(Copy(codeLine, x, 1) = '=') Then Begin
				x := x + 1;
				initVal := CutAtNextDelimit(codeLine, x, ' #/');
			END;

			GetToNextChar(codeLine, x);
			nextChar := NextCharInLine(codeLine, x);  // since there may be end-spaces, we need to check nextChar's length.  
			WHILE ((x <= Length(codeLine)) AND (nextChar <> '#') AND (nextChar <> '/') AND (Length(nextChar) > 0)) Do Begin
				extCount := extCount + 1;
				exten[extCount] := CutAtNextDelimit(codeLine, x, ' =,#/');
				GetToNextChar(codeLine, x);
				If(Copy(codeLine, x, 1) = '=') Then Begin
					x := x + 1;
					extVal[extCount] := CutAtNextDelimit(codeLine, x, ' ,#/');
				END;
				If(Copy(codeLine, x, 1) = ',') Then x := x + 1;     {get past the comma}
				nextChar := NextCharInLine(codeLine, x);
			End;

			{---------- Record Variable -------------}

			varCount := varCount + 1;      {not using a With..Do here because we don't wan't a conflict with existing variables.}
			symbol[varCount].name := name;
			symbol[varCount].stype := stype;
			symbol[varCount].initVal := initVal;
			symbol[varCount].extCount := extCount;
			FOR a := 1 TO extCount Do Begin
				symbol[varCount].exten[a] := exten[a];
				symbol[varCount].extVal[a] := extVal[a];
			End;

			If(initVal = 'null') Then Begin
				If(Length(name) > longName) Then longName := Length(name);
			End
			ELSE If(Length(name) + Length(initVal) + 1 > longName) Then longName := Length(name) + Length(initVal) + 1;
		END
		Else If(((nextChar = '#') Or (nextChar = '/')) AND ((IsCommentLine(codeLine) = 0) OR (addSymComLines = 0)) ) THEN StoreComment(Trim(codeLine));
	End;
END;

Procedure ReadFileData;
Var
	x : integer;
	section, codeWord, value : string;
Begin
	AssignFile(file1, ExtractFilePath(ParamStr(0)) + 'data.dat');
	Reset(file1);
	WHILE not EOF(file1) Do Begin
		ReadLn(file1, codeLine);
		If((NextCharInLine(codeLine, 1) <> '#') AND (CheckForBlankLine(codeLine, 1) = 0)) Then Begin
			If(Pos(':', codeLine) > 0) Then
				section := codeLine
			ELSE Begin
				If(section = 'keywords:') Then Begin
					numKeys := numKeys + 1;
					keyWord[numKeys] := codeLine;
				End
				ELSE If(section = 'settings:') Then Begin
					x := 1;
					codeWord := CutAtNextDelimit(codeLine, x, '=');
					x := x + 1;
					value := Copy(codeLine, x, Length(codeLine) - x + 1);
					If(codeWord = 'backup') Then makeBackup := StrToInt(value)
					Else if(codeWord = 'breakBrack') Then breakBrack := StrToInt(value)
					Else if(codeWord = 'brackBreak') Then brackBreak := StrToInt(value)
					Else if(codeWord = 'breakEndBrack') Then breakEndBrack := StrToInt(value)
					Else if(codeWord = 'endBrackBreak') Then endBrackBreak := StrToInt(value)
					Else if(codeWord = 'multiStat') Then multiStat := StrToInt(value)
					Else if(codeWord = 'maxLineLen') Then maxLineLen := StrToInt(value)
					Else if(codeWord = 'indentBlocks') Then indentBlocks := StrToInt(value)
					Else if(codeWord = 'indentMsgs') Then indentMsgs := StrToInt(value)
					Else if(codeWord = 'msgBreak') Then msgBreak := StrToInt(value)
					Else if(codeWord = 'indentReturn') Then indentReturn := StrToInt(value)
					Else if(codeWord = 'breakReturn') Then breakReturn := StrToInt(value)
					Else if(codeWord = 'addSymComLines') Then addSymComLines := StrToInt(value)
					Else if(codeWord = 'addCodeComLines') Then addCodeComLines := StrToInt(value)
					Else if(codeWord = 'breakSymbols') Then breakSymbols := StrToInt(value)
					Else if(codeWord = 'maxExpLen') Then maxExpLen := StrToInt(value)
					Else if(codeWord = 'codeComLine') Then codeComLine := value
					Else if(codeWord = 'symComLine') Then symComLine := value;
				END;
			END;
		END;
	End;
	CloseFile(file1);
END;

Procedure StoreComment(codeLine:string);
Begin
	numComs := numComs + 1;
	comment[numComs] := codeLine;
END;

{------------------- Main Program Body -------------------------------------}

Begin
	blockLvl := 0;
	numComs := 0;
	longName := 0;
	lineNum := 0;
	numKeys := 0;
	gd1 := 0;
	gd2 := 0;
	lineLen := 0;

	ReadFileData;
	OpenCogs;
	ParseDesc;
	ParseFlags;
	ParseSymbols;
	OutputSymbols;
	ParseDesc;
	ParseCode;
	CloseCogs;
END.
